@@ -2,14 +2,16 @@ import logging
 import os
 import sys
 from abc import ABC, abstractmethod
-from contextlib import contextmanager
+from contextlib import ExitStack, contextmanager
 from typing import Callable, Dict, Iterator, List, Optional, Sequence, Tuple, Union
 
 import matplotlib
 import numpy as np
 from tqdm.auto import tqdm
+from typing_extensions import TypedDict
 
 from qcodes import config
+from qcodes.dataset.data_set import res_type
 from qcodes.dataset.data_set_protocol import DataSetProtocol
 from qcodes.dataset.descriptions.detect_shapes import detect_shape_of_measurement
 from qcodes.dataset.descriptions.versioning.rundescribertypes import Shapes
@@ -17,7 +19,11 @@ from qcodes.dataset.experiment_container import Experiment
 from qcodes.dataset.measurements import Measurement
 from qcodes.dataset.plotting import plot_dataset
 from qcodes.instrument.parameter import _BaseParameter
-from qcodes.utils.threading import process_params_meas
+from qcodes.utils.threading import (
+    SequentialParamsCaller,
+    ThreadPoolParamsCaller,
+    process_params_meas,
+)
 
 ActionsT = Sequence[Callable[[], None]]
 
@@ -32,31 +38,40 @@ AxesTupleListWithDataSet = Tuple[
     List[matplotlib.axes.Axes],
     List[Optional[matplotlib.colorbar.Colorbar]],
 ]
+MultiAxesTupleListWithDataSet = Tuple[
+    Tuple[DataSetProtocol, ...],
+    Tuple[List[matplotlib.axes.Axes], ...],
+    Tuple[List[Optional[matplotlib.colorbar.Colorbar]], ...],
+]
 
 LOG = logging.getLogger(__name__)
 
 
+class ParameterGroup(TypedDict):
+    params: Tuple[ParamMeasT, ...]
+    meas_name: str
+    measured_params: List[res_type]
+
+
 class UnsafeThreadingException(Exception):
     pass
 
 
 def _register_parameters(
-        meas: Measurement,
-        param_meas: Sequence[ParamMeasT],
-        setpoints: Optional[Sequence[_BaseParameter]] = None,
-        shapes: Shapes = None
-        ) -> None:
+    meas: Measurement,
+    param_meas: Sequence[ParamMeasT],
+    setpoints: Optional[Sequence[_BaseParameter]] = None,
+    shapes: Shapes = None,
+) -> None:
     for parameter in param_meas:
         if isinstance(parameter, _BaseParameter):
-            meas.register_parameter(parameter,
-                                    setpoints=setpoints)
+            meas.register_parameter(parameter, setpoints=setpoints)
     meas.set_shapes(shapes=shapes)
 
 
 def _register_actions(
-        meas: Measurement,
-        enter_actions: ActionsT,
-        exit_actions: ActionsT) -> None:
+    meas: Measurement, enter_actions: ActionsT, exit_actions: ActionsT
+) -> None:
     for action in enter_actions:
         # this omits the possibility of passing
         # argument to enter and exit actions.
@@ -66,9 +81,7 @@ def _register_actions(
         meas.add_after_run(action, ())
 
 
-def _set_write_period(
-        meas: Measurement,
-        write_period: Optional[float] = None) -> None:
+def _set_write_period(meas: Measurement, write_period: Optional[float] = None) -> None:
     if write_period is not None:
         meas.write_period = write_period
 
@@ -141,19 +154,15 @@ def do0d(
     except TypeError:
         LOG.exception(
             f"Could not detect shape of {measured_parameters} "
-            f"falling back to unknown shape.")
+            f"falling back to unknown shape."
+        )
         shapes = None
 
     _register_parameters(meas, param_meas, shapes=shapes)
     _set_write_period(meas, write_period)
 
     with meas.run() as datasaver:
-        datasaver.add_result(
-            *process_params_meas(
-                param_meas,
-                use_threads=use_threads
-            )
-        )
+        datasaver.add_result(*process_params_meas(param_meas, use_threads=use_threads))
         dataset = datasaver.dataset
 
     return _handle_plotting(dataset, do_plot)
@@ -234,30 +243,36 @@ def do1d(
         param for param in param_meas if isinstance(param, _BaseParameter)
     )
     try:
-        loop_shape = tuple(1 for _ in additional_setpoints) + (num_points,)
-        shapes: Shapes = detect_shape_of_measurement(
-            measured_parameters,
-            loop_shape
-        )
+        loop_shape = (num_points,) + tuple(1 for _ in additional_setpoints)
+        shapes: Shapes = detect_shape_of_measurement(measured_parameters, loop_shape)
     except TypeError:
         LOG.exception(
             f"Could not detect shape of {measured_parameters} "
-            f"falling back to unknown shape.")
+            f"falling back to unknown shape."
+        )
         shapes = None
 
     _register_parameters(meas, all_setpoint_params)
-    _register_parameters(meas, param_meas, setpoints=all_setpoint_params,
-                         shapes=shapes)
+    _register_parameters(meas, param_meas, setpoints=all_setpoint_params, shapes=shapes)
     _set_write_period(meas, write_period)
     _register_actions(meas, enter_actions, exit_actions)
 
     original_delay = param_set.post_delay
     param_set.post_delay = delay
 
+    if use_threads is None:
+        use_threads = config.dataset.use_threads
+
+    param_meas_caller = (
+        ThreadPoolParamsCaller(*param_meas)
+        if use_threads
+        else SequentialParamsCaller(*param_meas)
+    )
+
     # do1D enforces a simple relationship between measured parameters
     # and set parameters. For anything more complicated this should be
     # reimplemented from scratch
-    with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver:
+    with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver, param_meas_caller as call_param_meas:
         dataset = datasaver.dataset
         additional_setpoints_data = process_params_meas(additional_setpoints)
         setpoints = np.linspace(start, stop, num_points)
@@ -270,9 +285,7 @@ def do1d(
         for set_point in tqdm(setpoints, disable=not show_progress):
             param_set.set(set_point)
             datasaver.add_result(
-                (param_set, set_point),
-                *process_params_meas(param_meas, use_threads=use_threads),
-                *additional_setpoints_data
+                (param_set, set_point), *call_param_meas(), *additional_setpoints_data
             )
 
     param_set.post_delay = original_delay
@@ -374,26 +387,22 @@ def do2d(
         param_set2,
     ) + tuple(s for s in additional_setpoints)
 
-    measured_parameters = tuple(param for param in param_meas
-                                if isinstance(param, _BaseParameter))
+    measured_parameters = tuple(
+        param for param in param_meas if isinstance(param, _BaseParameter)
+    )
 
     try:
-        loop_shape = tuple(
-            1 for _ in additional_setpoints
-        ) + (num_points1, num_points2)
-        shapes: Shapes = detect_shape_of_measurement(
-            measured_parameters,
-            loop_shape
-        )
+        loop_shape = (num_points1, num_points2) + tuple(1 for _ in additional_setpoints)
+        shapes: Shapes = detect_shape_of_measurement(measured_parameters, loop_shape)
     except TypeError:
         LOG.exception(
             f"Could not detect shape of {measured_parameters} "
-            f"falling back to unknown shape.")
+            f"falling back to unknown shape."
+        )
         shapes = None
 
     _register_parameters(meas, all_setpoint_params)
-    _register_parameters(meas, param_meas, setpoints=all_setpoint_params,
-                         shapes=shapes)
+    _register_parameters(meas, param_meas, setpoints=all_setpoint_params, shapes=shapes)
     _set_write_period(meas, write_period)
     _register_actions(meas, enter_actions, exit_actions)
 
@@ -403,7 +412,16 @@ def do2d(
     param_set1.post_delay = delay1
     param_set2.post_delay = delay2
 
-    with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver:
+    if use_threads is None:
+        use_threads = config.dataset.use_threads
+
+    param_meas_caller = (
+        ThreadPoolParamsCaller(*param_meas)
+        if use_threads
+        else SequentialParamsCaller(*param_meas)
+    )
+
+    with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver, param_meas_caller as call_param_meas:
         dataset = datasaver.dataset
         additional_setpoints_data = process_params_meas(additional_setpoints)
         setpoints1 = np.linspace(start1, stop1, num_points1)
@@ -422,9 +440,7 @@ def do2d(
             # updates
             sys.stdout.flush()
             sys.stderr.flush()
-            for set_point2 in tqdm(setpoints2,
-                                   disable=not show_progress,
-                                   leave=False):
+            for set_point2 in tqdm(setpoints2, disable=not show_progress, leave=False):
                 # skip first inner set point if `set_before_sweep`
                 if set_point2 == start2 and set_before_sweep:
                     pass
@@ -434,8 +450,8 @@ def do2d(
                 datasaver.add_result(
                     (param_set1, set_point1),
                     (param_set2, set_point2),
-                    *process_params_meas(param_meas, use_threads=use_threads),
-                    *additional_setpoints_data
+                    *call_param_meas(),
+                    *additional_setpoints_data,
                 )
 
             for action in after_inner_actions:
@@ -485,6 +501,14 @@ class AbstractSweep(ABC):
         """
         pass
 
+    @property
+    @abstractmethod
+    def post_actions(self) -> ActionsT:
+        """
+        actions to be performed after setting param to its setpoint.
+        """
+        pass
+
 
 class LinSweep(AbstractSweep):
     """
@@ -505,12 +529,14 @@ class LinSweep(AbstractSweep):
         stop: float,
         num_points: int,
         delay: float = 0,
+        post_actions: ActionsT = (),
     ):
         self._param = param
         self._start = start
         self._stop = stop
         self._num_points = num_points
         self._delay = delay
+        self._post_actions = post_actions
 
     def get_setpoints(self) -> np.ndarray:
         """
@@ -531,6 +557,10 @@ class LinSweep(AbstractSweep):
     def num_points(self) -> int:
         return self._num_points
 
+    @property
+    def post_actions(self) -> ActionsT:
+        return self._post_actions
+
 
 class LogSweep(AbstractSweep):
     """
@@ -551,12 +581,14 @@ class LogSweep(AbstractSweep):
         stop: float,
         num_points: int,
         delay: float = 0,
+        post_actions: ActionsT = (),
     ):
         self._param = param
         self._start = start
         self._stop = stop
         self._num_points = num_points
         self._delay = delay
+        self._post_actions = post_actions
 
     def get_setpoints(self) -> np.ndarray:
         """
@@ -577,9 +609,13 @@ class LogSweep(AbstractSweep):
     def num_points(self) -> int:
         return self._num_points
 
+    @property
+    def post_actions(self) -> ActionsT:
+        return self._post_actions
+
 
 def dond(
-    *params: Union[AbstractSweep, ParamMeasT],
+    *params: Union[AbstractSweep, Union[ParamMeasT, Sequence[ParamMeasT]]],
     write_period: Optional[float] = None,
     measurement_name: str = "",
     exp: Optional[Experiment] = None,
@@ -590,7 +626,7 @@ def dond(
     use_threads: Optional[bool] = None,
     additional_setpoints: Sequence[ParamMeasT] = tuple(),
     log_info: Optional[str] = None,
-) -> AxesTupleListWithDataSet:
+) -> Union[AxesTupleListWithDataSet, MultiAxesTupleListWithDataSet]:
     """
     Perform n-dimentional scan from slowest (first) to the fastest (last), to
     measure m measurement parameters. The dimensions should be specified
@@ -606,6 +642,15 @@ def dond(
                 LinSweep(param_set_n, start_n, stop_n, num_points_n, delay_n),
                 param_meas_1, param_meas_2, ..., param_meas_m
 
+            If multiple DataSets creation is needed, measurement parameters should
+            be grouped, so one dataset will be created for each group. e.g.:
+
+            .. code-block::
+
+                LinSweep(param_set_1, start_1, stop_1, num_points_1, delay_1), ...,
+                LinSweep(param_set_n, start_n, stop_n, num_points_n, delay_n),
+                [param_meas_1, param_meas_2], ..., [param_meas_m]
+
         write_period: The time after which the data is actually written to the
             database.
         measurement_name: Name of the measurement. This will be passed down to
@@ -628,26 +673,28 @@ def dond(
             the measurement but not scanned/swept-over.
         log_info: Message that is logged during the measurement. If None a default
             message is used.
+
+        Returns:
+            A tuple of QCoDeS DataSet, Matplotlib axis, Matplotlib colorbar. If
+            more than one group of measurement parameters is supplied, the output
+            will be a tuple of tuple(QCoDeS DataSet), tuple(Matplotlib axis),
+            tuple(Matplotlib colorbar), in which each element of each sub-tuple
+            belongs to one group, and the order of elements is the order of
+            the supplied groups.
     """
     if do_plot is None:
         do_plot = config.dataset.dond_plot
     if show_progress is None:
         show_progress = config.dataset.dond_show_progress
 
-    meas = Measurement(name=measurement_name, exp=exp)
-    if log_info is not None:
-        meas._extra_log_info = log_info
-    else:
-        meas._extra_log_info = "Using 'qcodes.utils.dataset.doNd.dond'"
-
     def _parse_dond_arguments(
-        *params: Union[AbstractSweep, ParamMeasT]
-    ) -> Tuple[List[AbstractSweep], List[ParamMeasT]]:
+        *params: Union[AbstractSweep, Union[ParamMeasT, Sequence[ParamMeasT]]]
+    ) -> Tuple[List[AbstractSweep], List[Union[ParamMeasT, Sequence[ParamMeasT]]]]:
         """
         Parse supplied arguments into sweep objects and measurement parameters.
         """
         sweep_instances: List[AbstractSweep] = []
-        params_meas: List[ParamMeasT] = []
+        params_meas: List[Union[ParamMeasT, Sequence[ParamMeasT]]] = []
         for par in params:
             if isinstance(par, AbstractSweep):
                 sweep_instances.append(par)
@@ -671,13 +718,15 @@ def dond(
         s for s in additional_setpoints
     )
 
-    measured_parameters = tuple(
-        par for par in params_meas if isinstance(par, _BaseParameter)
-    )
+    (
+        all_meas_parameters,
+        grouped_parameters,
+        measured_parameters,
+    ) = _extract_paramters_by_type_and_group(measurement_name, params_meas)
 
     try:
-        loop_shape = tuple(1 for _ in additional_setpoints) + tuple(
-            sweep.num_points for sweep in sweep_instances
+        loop_shape = tuple(sweep.num_points for sweep in sweep_instances) + tuple(
+            1 for _ in additional_setpoints
         )
         shapes: Shapes = detect_shape_of_measurement(measured_parameters, loop_shape)
     except TypeError:
@@ -686,41 +735,174 @@ def dond(
             f"falling back to unknown shape."
         )
         shapes = None
-
-    _register_parameters(meas, all_setpoint_params)
-    _register_parameters(
-        meas, params_meas, setpoints=all_setpoint_params, shapes=shapes
+    meas_list = _create_measurements(
+        all_setpoint_params,
+        enter_actions,
+        exit_actions,
+        exp,
+        grouped_parameters,
+        shapes,
+        write_period,
+        log_info,
     )
-    _set_write_period(meas, write_period)
-    _register_actions(meas, enter_actions, exit_actions)
 
     original_delays: Dict[_BaseParameter, float] = {}
     params_set: List[_BaseParameter] = []
+    post_actions: List[ActionsT] = []
     for sweep in sweep_instances:
         original_delays[sweep.param] = sweep.param.post_delay
         sweep.param.post_delay = sweep.delay
         params_set.append(sweep.param)
+        post_actions.append(sweep.post_actions)
+
+    datasets = []
+    plots_axes = []
+    plots_colorbar = []
+    if use_threads is None:
+        use_threads = config.dataset.use_threads
+
+    params_meas_caller = (
+        ThreadPoolParamsCaller(*all_meas_parameters)
+        if use_threads
+        else SequentialParamsCaller(*all_meas_parameters)
+    )
 
     try:
-        with _catch_keyboard_interrupts() as interrupted, meas.run() as datasaver:
-            dataset = datasaver.dataset
+        with _catch_keyboard_interrupts() as interrupted, ExitStack() as stack, params_meas_caller as call_params_meas:
+            datasavers = [stack.enter_context(measure.run()) for measure in meas_list]
             additional_setpoints_data = process_params_meas(additional_setpoints)
+            previous_setpoints = np.empty(len(sweep_instances))
             for setpoints in tqdm(nested_setpoints, disable=not show_progress):
+
+                active_actions = _select_active_actions(
+                    post_actions, setpoints, previous_setpoints
+                )
+                previous_setpoints = setpoints
+
                 param_set_list = []
-                param_value_pairs = zip(params_set[::-1], setpoints[::-1])
-                for setpoint_param, setpoint in param_value_pairs:
+                param_value_action = zip(params_set, setpoints, active_actions)
+                for setpoint_param, setpoint, action in param_value_action:
                     setpoint_param(setpoint)
                     param_set_list.append((setpoint_param, setpoint))
-                datasaver.add_result(
-                    *param_set_list,
-                    *process_params_meas(params_meas, use_threads=use_threads),
-                    *additional_setpoints_data,
-                )
+                    for act in action:
+                        act()
+
+                meas_value_pair = call_params_meas()
+                for group in grouped_parameters.values():
+                    group["measured_params"] = []
+                    for measured in meas_value_pair:
+                        if measured[0] in group["params"]:
+                            group["measured_params"].append(measured)
+                for ind, datasaver in enumerate(datasavers):
+                    datasaver.add_result(
+                        *param_set_list,
+                        *grouped_parameters[f"group_{ind}"]["measured_params"],
+                        *additional_setpoints_data,
+                    )
+
     finally:
         for parameter, original_delay in original_delays.items():
             parameter.post_delay = original_delay
 
-    return _handle_plotting(dataset, do_plot, interrupted())
+        for datasaver in datasavers:
+            ds, plot_axis, plot_color = _handle_plotting(
+                datasaver.dataset, do_plot, interrupted()
+            )
+            datasets.append(ds)
+            plots_axes.append(plot_axis)
+            plots_colorbar.append(plot_color)
+
+    if len(grouped_parameters) == 1:
+        return datasets[0], plots_axes[0], plots_colorbar[0]
+    else:
+        return tuple(datasets), tuple(plots_axes), tuple(plots_colorbar)
+
+
+def _select_active_actions(
+    actions: Sequence[ActionsT], setpoints: np.ndarray, previous_setpoints: np.ndarray
+) -> List[ActionsT]:
+    """
+    Select ActionT (Sequence[Callable]) from a Sequence of ActionsT if
+    the corresponding setpoint has changed. Otherwise select an empty Sequence.
+    """
+    actions_list: List[ActionsT] = [()] * len(setpoints)
+    for ind, (new_setpoint, old_setpoint) in enumerate(
+        zip(setpoints, previous_setpoints)
+    ):
+        if new_setpoint != old_setpoint:
+            actions_list[ind] = actions[ind]
+    return actions_list
+
+
+def _create_measurements(
+    all_setpoint_params: Sequence[_BaseParameter],
+    enter_actions: ActionsT,
+    exit_actions: ActionsT,
+    exp: Optional[Experiment],
+    grouped_parameters: Dict[str, ParameterGroup],
+    shapes: Shapes,
+    write_period: Optional[float],
+    log_info: Optional[str],
+) -> Tuple[Measurement, ...]:
+    meas_list: List[Measurement] = []
+    if log_info is not None:
+        _extra_log_info = log_info
+    else:
+        _extra_log_info = "Using 'qcodes.utils.dataset.doNd.dond'"
+    for group in grouped_parameters.values():
+        meas_name = group["meas_name"]
+        meas_params = group["params"]
+        meas = Measurement(name=meas_name, exp=exp)
+        meas._extra_log_info = _extra_log_info
+        _register_parameters(meas, all_setpoint_params)
+        _register_parameters(
+            meas, meas_params, setpoints=all_setpoint_params, shapes=shapes
+        )
+        _set_write_period(meas, write_period)
+        _register_actions(meas, enter_actions, exit_actions)
+        meas_list.append(meas)
+    return tuple(meas_list)
+
+
+def _extract_paramters_by_type_and_group(
+    measurement_name: str,
+    params_meas: Sequence[Union[ParamMeasT, Sequence[ParamMeasT]]],
+) -> Tuple[
+    Tuple[ParamMeasT, ...], Dict[str, ParameterGroup], Tuple[_BaseParameter, ...]
+]:
+    measured_parameters: List[_BaseParameter] = []
+    all_meas_parameters: List[ParamMeasT] = []
+    single_group: List[ParamMeasT] = []
+    multi_group: List[Sequence[ParamMeasT]] = []
+    grouped_parameters: Dict[str, ParameterGroup] = {}
+    for param in params_meas:
+        if not isinstance(param, Sequence):
+            single_group.append(param)
+            all_meas_parameters.append(param)
+            if isinstance(param, _BaseParameter):
+                measured_parameters.append(param)
+        elif not isinstance(param, str):
+            multi_group.append(param)
+            for nested_param in param:
+                all_meas_parameters.append(nested_param)
+                if isinstance(nested_param, _BaseParameter):
+                    measured_parameters.append(nested_param)
+    if single_group:
+        pg: ParameterGroup = {
+            "params": tuple(single_group),
+            "meas_name": measurement_name,
+            "measured_params": [],
+        }
+        grouped_parameters["group_0"] = pg
+    if multi_group:
+        for index, par in enumerate(multi_group):
+            pg = {
+                "params": tuple(par),
+                "meas_name": measurement_name,
+                "measured_params": [],
+            }
+            grouped_parameters[f"group_{index}"] = pg
+    return tuple(all_meas_parameters), grouped_parameters, tuple(measured_parameters)
 
 
 def _handle_plotting(
@@ -769,16 +951,16 @@ def plot(
     sample_name = data.sample_name
     storage_dir = os.path.join(mainfolder, experiment_name, sample_name)
     os.makedirs(storage_dir, exist_ok=True)
-    png_dir = os.path.join(storage_dir, 'png')
-    pdf_dif = os.path.join(storage_dir, 'pdf')
+    png_dir = os.path.join(storage_dir, "png")
+    pdf_dif = os.path.join(storage_dir, "pdf")
     os.makedirs(png_dir, exist_ok=True)
     os.makedirs(pdf_dif, exist_ok=True)
     for i, ax in enumerate(axes):
         if save_pdf:
-            full_path = os.path.join(pdf_dif, f'{dataid}_{i}.pdf')
+            full_path = os.path.join(pdf_dif, f"{dataid}_{i}.pdf")
             ax.figure.savefig(full_path, dpi=500)
         if save_png:
-            full_path = os.path.join(png_dir, f'{dataid}_{i}.png')
+            full_path = os.path.join(png_dir, f"{dataid}_{i}.png")
             ax.figure.savefig(full_path, dpi=500)
     res = data, axes, cbs
     return res
